#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naFile.py
=========

A containter module for the mixin base class NAFile that is subclassed
for individual FFIs. Each FFI class is held in an individual file.

"""
   
# 08/05/04 updated by selatham for bug fixes and new write methods

# Imports from python standard library
import sys
import time
import re
import StringIO

# Imports from nappy package
import nappy.na_file.na_core
import nappy.utils.text_parser
import nappy.utils.common_utils
import nappy.na_error

default_delimiter = nappy.utils.getDefault("default_delimiter")
default_float_format = nappy.utils.getDefault("default_float_format")
getAnnotation = nappy.utils.common_utils.getAnnotation
wrapLine = nappy.utils.common_utils.annotateLine
wrapLines = nappy.utils.common_utils.annotateLines
stripQuotes = nappy.utils.common_utils.stripQuotes


class NAFile(nappy.na_file.na_core.NACore):

    """
    NAFile class is a sub-class of NACore abstract classes.
    NAFile is also an abstract class and should not be called directly.
    
    NAFile holds all the methods that are common to either all or more than
    one NASA Ames FFI class. These methods set up the main read and write
    functionality for the NASA Ames format.

    When a sub-class of NAFile is called with a read ('r' - default) 
    mode the header in the file is automatically read. To read the data
    section the user must call the 'readData' method.
      
    When a sub-class of NAFile is called with the write ('w') mode
    then the file is opened but nothing is written. The user must then 
    send an 'na_dict' object to the 'write' method to write the output.
    The output file is then flushed to ensure the data is written even if 
    the user forgets to close it.
    """

    def __init__(self, filename, mode="r", na_dict={}): 
        """
        Initialization of class, decides if user wishes to read or write
        NASA Ames file.
        """
        nappy.na_file.na_core.NACore.__init__(self)
        self.filename = filename
        self._open(mode)
        self.mode = mode
        self.na_dict = na_dict

        if self.mode == "r":
            self._normalized_X = True
            self.readHeader()
        elif self.mode == "w":
            # Self flag to check if data written
            self.data_written = False
        else:
            raise "Unknown file mode '%s'." % self.mode

        
    def _open(self, mode):
        "Wrapper to builtin open file function."
        self.file = open(self.filename, mode)
        self.is_open = True

    def write(self, delimiter=default_delimiter, float_format=default_float_format,
              annotation=False, no_header=False):
        """
        Writes an na_dict to the file and then flushes it to ensure data not 
        being buffered.
        If annotation is True then add annotation column to left of file.
        If no_header is True then suppress writing the header and only write the data section. 
        """ 
        self.delimiter = delimiter
        self.float_format = float_format
        self.format = float_format + delimiter
        self.annotation = annotation
        
        # Raise errors if dangerous behaviour
        if self.mode != "w":
            raise Exception("WARNING: Cannot write to read-only file. Can only write to NA file object when mode='w'.")

        if self.data_written == True:
            raise Exception("WARNING: Cannot write multiple NASA Ames dictionaries to a single file. Please open a new NASA Ames file instance to write new data to.")

        if self.is_open == False:
            raise Exception("WARNING: NASA Ames file instance is closed and cannot be written to.")
   
        # Parse na_dict then write header and data
        self._parseDictionary()
        self.header = StringIO.StringIO()

        if no_header == False:
            self.writeHeader()

        self.writeData()
        self.file.flush()
        
        # Set flag to make sure cannot try and write more data
        self.data_written = True
         
    def close(self):
        "Wrapper to builtin close file function."
        self.file.close()
        self.is_open = False

    def _parseDictionary(self):
        """
        Parser for the optional na_dict argument containing a dictionary
        of NASA Ames internal variables. These are saved as instance attributes
        with the name used in the NASA Ames documentation.
        """
        for i in self.na_dict.keys():
            setattr(self, i, self.na_dict[i])

    def _readTopLine(self):
        """
        Reads number of header lines and File Format Index from top line.
        Also assigns a value to NIV for the number of independent variables
        based on the first character in the FFI.

        Returns NLHEAD and FFI in a tuple.
        """
        (self.NLHEAD, self.FFI) = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), 2, int)
        self.NIV = int(self.FFI/1000)
        return (self.NLHEAD, self.FFI)

    def _readLines(self, nlines):
        "Reads nlines lines from a file and returns them in a list."
        lines = []
        for i in range(nlines):
            lines.append(self.file.readline().strip())
        return lines

    def _checkForBlankLines(self, datalines):
        """
        Searches for empty lines in the middle of the data section and raises
        as error if found. It ignores empty lines at the end of the file but
        strips them out before returning a list of lines for reading.
        """
        empties = None
        count = 0
        rtlines = []
        for line in datalines:
            if line.strip() == "":
                empties = 1
            else:
                if empties == 1:   # If data line found after empty line then raise
                    raise Exception("Empty line found in data section at line: " + `count`)
                else:
                    rtlines.append(line)
            count = count + 1
        return rtlines


    def _readCommonHeader(self):
        """
        Reads the header section common to all NASA Ames files.
        """
        self._readTopLine()
        self.ONAME = nappy.utils.text_parser.readItemFromLine(self.file.readline(), str)
        self.ORG = nappy.utils.text_parser.readItemFromLine(self.file.readline(), str)
        self.SNAME = nappy.utils.text_parser.readItemFromLine(self.file.readline(), str)
        self.MNAME = nappy.utils.text_parser.readItemFromLine(self.file.readline(), str)
        (self.IVOL, self.NVOL) = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), 2, int)
        dates = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), 6, int)
        (self.DATE, self.RDATE) = (dates[:3], dates[3:])

    def _writeCommonHeader(self):
        """
        Writes the header section common to all NASA Ames files.
        """
        #Line 1 if often overwritten at _fixHeaderLength
        self.header.write(wrapLine("NLHEAD_FFI", self.annotation, self.delimiter, "%d%s%d\n" % (self.NLHEAD, self.delimiter, self.FFI)))
        self.header.write(getAnnotation("ONAME", self.annotation, delimiter = self.delimiter) + stripQuotes(self.ONAME) + "\n")
        self.header.write(getAnnotation("ORG", self.annotation, delimiter = self.delimiter) + stripQuotes(self.ORG) + "\n")
        self.header.write(getAnnotation("SNAME", self.annotation, delimiter = self.delimiter) + stripQuotes(self.SNAME) + "\n")
        self.header.write(getAnnotation("MNAME", self.annotation, delimiter = self.delimiter) + stripQuotes(self.MNAME) + "\n")
        self.header.write(wrapLine("IVOL_NVOL", self.annotation, self.delimiter, "%d%s%d\n" % (self.IVOL, self.delimiter, self.NVOL)))
        line = "%d %d %d%s%d %d %d\n" % (self.DATE[0], self.DATE[1], self.DATE[2], self.delimiter, self.RDATE[0], self.RDATE[1], self.RDATE[2])
        self.header.write(wrapLine("DATE_RDATE", self.annotation, self.delimiter, line))

    def _readVariablesHeaderSection(self):
        """
        Reads the variables section of the header.
        Assumes we are at the right point in the file.
        """
        self.NV = nappy.utils.text_parser.readItemFromLine(self.file.readline(), int)
        self.VSCAL = nappy.utils.text_parser.readItemsFromUnknownLines(self.file, self.NV, float)
        self.VMISS = nappy.utils.text_parser.readItemsFromUnknownLines(self.file, self.NV, float)
        self.VNAME = nappy.utils.text_parser.readItemsFromLines(self._readLines(self.NV), self.NV, str)

    def _writeVariablesHeaderSection(self):
        """
        Writes the variables section of the header.
        Assumes we are at the right point in the file.
        """
        self.header.write(wrapLine("NV", self.annotation, self.delimiter, "%d\n" % self.NV))
        self.header.write(wrapLine("VSCAL", self.annotation, self.delimiter, (("%s" + self.delimiter) * (self.NV - 1) + "%s\n") % tuple(self.VSCAL)))
        self.header.write(wrapLine("VMISS", self.annotation, self.delimiter, (("%s" + self.delimiter) * (self.NV - 1) + "%s\n") % tuple(self.VMISS)))
        self.header.write(wrapLines("VNAME", self.annotation, self.delimiter, "%s\n" * self.NV % tuple(self.VNAME)))

    def _readAuxVariablesHeaderSection(self):
        """
        Reads the auxiliary variables section of the header.
        Assumes we are at the right point in the file.
        """
        self.NAUXV = nappy.utils.text_parser.readItemFromLine(self.file.readline(), int)
        if self.NAUXV > 0:        
            self.ASCAL = nappy.utils.text_parser.readItemsFromUnknownLines(self.file, self.NAUXV, float)
            self.AMISS = nappy.utils.text_parser.readItemsFromUnknownLines(self.file, self.NAUXV, float)
            self.ANAME = nappy.utils.text_parser.readItemsFromLines(self._readLines(self.NAUXV), self.NAUXV, str)

    def _writeAuxVariablesHeaderSection(self):
        """
        Writes the auxiliary variables section of the header.
        Assumes we are at the right point in the file.
        """
        self.header.write(wrapLine("NAUXV", self.annotation, self.delimiter, "%d\n" % self.NAUXV))
        if self.NAUXV > 0:
            line = (("%s" + self.delimiter) * (self.NAUXV - 1) + "%s\n")  % tuple(self.ASCAL)
            self.header.write(wrapLine("ASCAL", self.annotation, self.delimiter, line))
            line = (("%s" + self.delimiter) * (self.NAUXV - 1) + "%s\n")  % tuple(self.AMISS)
            self.header.write(wrapLine("AMISS", self.annotation, self.delimiter, line))
            line = "%s\n" * self.NAUXV % tuple(self.ANAME)
            self.header.write(wrapLines("ANAME", self.annotation, self.delimiter, line))

    def _readCharAuxVariablesHeaderSection(self):
        """
        Reads the character-encoded auxiliary variables section of the header.
        Assumes we are at the right point in the file.
        """
        self.NAUXV = nappy.utils.text_parser.readItemFromLine(self.file.readline(), int)
        self.NAUXC = nappy.utils.text_parser.readItemFromLine(self.file.readline(), int)
        nonCharAuxVars = self.NAUXV - self.NAUXC
        if self.NAUXV > 0:
            self.ASCAL = nappy.utils.text_parser.readItemsFromUnknownLines(self.file, nonCharAuxVars, float)
            self.AMISS = nappy.utils.text_parser.readItemsFromUnknownLines(self.file, nonCharAuxVars, float)
            self.LENA = nappy.utils.text_parser.readItemsFromUnknownLines(self.file, self.NAUXC, int)
            for i in range(nonCharAuxVars):
                self.LENA.insert(0, None)
            self.AMISS = self.AMISS + nappy.utils.text_parser.readItemsFromUnknownLines(self.file, self.NAUXC, str)    
            self.ANAME = nappy.utils.text_parser.readItemsFromLines(self._readLines(self.NAUXV), self.NAUXV, str)        
            
    def _readComments(self):
        """
        Reads the special and normal comments sections.
        Assumes we are at the right point in the file.
        """        
        self.NSCOML = nappy.utils.text_parser.readItemFromLine(self.file.readline(), int)
        self._readSpecialComments()
        self.NNCOML = nappy.utils.text_parser.readItemFromLine(self.file.readline(), int)
        self._readNormalComments()

    def _writeComments(self):
        """
        Writes the special and normal comments sections.
        Assumes we are at the right point in the file.
        """
        self.header.write(wrapLine("NSCOML", self.annotation, self.delimiter, "%d\n" % self.NSCOML))
        self.header.write(wrapLines("SCOM", self.annotation, self.delimiter, "%s\n" * self.NSCOML % tuple(self.SCOM)))
        self.header.write(wrapLine("NNCOML", self.annotation, self.delimiter, "%d\n" % self.NNCOML))
        self.header.write(wrapLines("NCOM", self.annotation, self.delimiter, "%s\n" * self.NNCOML % tuple(self.NCOM)))

    def _fixHeaderLength(self):
        """
        Takes the self.header StringIO object and counts the number of lines
        and corrects the NLHEAD value in the header line.
        Resets to start of self.header.
        """
        self.header.seek(0)
        lines = self.header.readlines()
        headlength = len(lines)
        lines[0] = wrapLine("NLHEAD_FFI", self.annotation, self.delimiter, "%d%s%d\n" % (headlength, self.delimiter, self.FFI))
        self.header = StringIO.StringIO("".join(lines))
        self.header.seek(0) 

    def _readSpecialComments(self):
        """
        Reads the special comments section.        
        Assumes that we are at the right point in the file and that NSCOML
        variable is known.
        """
        self.SCOM = self._readLines(self.NSCOML)
        return self.SCOM

    def _readNormalComments(self):
        """
        Reads the normal comments section.        
        Assumes that we are at the right point in the file and that NNCOML
        variable is known.
        """
        self.NCOM = self._readLines(self.NNCOML)
        return self.NCOM

    def readData(self):
        """
        Reads the data section of the file. This method actually calls a number
        of FFI specific methods to setup the data arrays (lists of lists) and
        read the various data sections.

        This method can be called directly by the user.
        """
        self._setupArrays()
        datalines = open(self.filename).readlines()[self.NLHEAD:]
        datalines = self._checkForBlankLines(datalines)

        # Set up loop over unbounded indpendent variable
        m = 0   # Unbounded independent variable mark        
        while len(datalines) > 0:
            datalines = self._readData1(datalines, m)
            datalines = self._readData2(datalines, m)
            m = m + 1
            



#!/usr/bin/env python
#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
cdms_to_na.py
=============

Holds the class CDMSToNA that converts a set of CDMS variables and global attributes.

"""

# Imports from python standard library
import sys
import os
import time
import string
import re

# Import from nappy package
from nappy.na_error import na_error
import nappy.utils
import nappy.utils.common_utils
import nappy.na_file.na_core


nc_to_na_map = utils.getConfigDict()["nc_to_na_map"]

# Import external packages (if available)
if sys.platform.find("win") > -1:
    raise na_error.NAPlatformError("Windows does not support CDMS. CDMS is required to convert to CDMS objects and NetCDF.")
try:
    import cdms, Numeric
except:
    raise Exception("Could not import third-party software. Nappy requires the CDMS and Numeric packages to be installed to convert to CDMS and NetCDF.")

cdms.setAutoBounds("off") 

compareAxes = cdms_utils.axis_utils.compareAxes
compareVariables = cdms_utils.var_utils.compareVariables
arrayToList = nappy.utils.list_manipulator.arrayToList
listOfListsCreator = nappy.utils.list_manipulator.listOfListsCreator
getBestName = cdms_utils.var_utils.getBestName
getMissingValue = cdms_utils.var_utils.getMissingValue
flatten2DTimeAxis = nappy.contrib.aircraft.aircraft_utils.flatten2DTimeAxis
modifyNADictCopy = nappy.utils.common_utils.modifyNADictCopy


def cdms2na(ncfile, na_file_names, naVars={}, variables=None, nFilesOnly="no", 
            rule=None, ffi="automatic", delimiter="    ", float_format="%g", 
            rules=None, sizeLimit=None):
    """
    Main conversion function that calls the appropriate classes and functions
    to write a NASA Ames file.
    """
    #print infilename, outfilenames, nFilesOnly, naVars, variables
    if type(na_file_names) == type("string"): 
        na_file_names = [na_file_names]
    
    # Get which NASA Ames internal variables are allowed to be overwritten in the output files (i.e. by user inputs)
    allowedOverwriteMetadata = ("DATE",  "RDATE", "ANAME", "MNAME",
           "ONAME", "ORG", "SNAME", "VNAME")
    arrayArgs=["DATE", "RDATE", "ANAME", "VNAME"]
    # ANAME[a] - array of 'a' x ANAME strings - aux var names
    # DATE (array of three) - UT date at which the data within the file starts 
    # MNAME - mission name 
    # ONAME - name of originator(s) 
    # ORG - org or affiliation of originator(s) 
    # RDATE (array of three) - date of data reduction or revision 
    # SNAME - source of measurement or model output VNAME[n] - array of 'n' x 
    # VNAME strings - var names.
    outputMessage=[]
    msg="Reading data from: %s\n" % infilename
    print msg
    outputMessage.append(msg)
    cdmsfile=cdms.open(infilename)
    globals=cdmsfile.attributes
    
    vars=[]
    if not variables:
        variables=cdmsfile.listvariables()
        #for var in cdmsfile.listvariables():
            #vars.append(cdmsfile(var))    
	    
    for variable in variables:
        varObj=cdmsfile(variable)
	# Deal with singleton variables
	if not hasattr(varObj, "rank"):
	        varMetadata=cdmsfile[variable].attributes
		varValue=varObj
		#print varMetadata, varValue, varMetadata.keys(), varMetadata._obj_.id
		varObj=cdms.createVariable(Numeric.array(varObj), id=getBestName(varMetadata).replace(" ", "_"), attributes=varMetadata)
		#print varObj, dir(varObj); sys.exit()
		varObj.value=varObj._data[0]
		#varObj.rank=0
		
	#print varObj, varObj.attributes	   		 
        vars.append(varObj)
	
    # Re-order variables if they have the attribute 'nasa_ames_var_number'
    orderedVars=[None]*1000
    otherVars=[]
    for var in vars:
        varMetadata=cdmsfile[var]
	if hasattr(varMetadata, "nasa_ames_var_number"):
	    num=varMetadata.nasa_ames_var_number
	    orderedVars[num]=var
	else:
	    otherVars.append(var)
    
    vars=[]
    for var in orderedVars:
        if var!=None:
	    vars.append(var)
	    
    vars=vars+otherVars
    
    builder=NAContentCollector(vars, globals, rule=rule, cdmsfile=cdmsfile)
    #print builder.na_dict["X"]
    builtNADicts=[[builder.na_dict, builder.varIDs]]
    if builder.varIDs==None:
        msg="\nNo files created after variables parsed."
        print msg
        outputMessage.append(msg)
        return outputMessage

    while len(builder.varBin)>0:
	builder=NAContentCollector(builder.varBin, globals, rule=rule, cdmsfile=cdmsfile)
	outputMessage=outputMessage+builder.outputMessage
        if builder.varIDs!=None:  builtNADicts.append([builder.na_dict, builder.varIDs])

    # Return only filenames if only want to know them now.
    ncount=1
    fileNames=[]
    if nFilesOnly=="yes": 
        for i in builtNADicts:
            if len(builtNADicts)==1:
	        suffix=""
	    else:
	        suffix="_%s" % ncount
	    nameparts=outfilenames[0].split(".")    
	    newname=(".".join(nameparts[:-1]))+suffix+"."+nameparts[-1]
	    fileNames.append(newname)
        ncount=ncount+1
	    
        return fileNames
	 	
    msg="\n%s files to write" % len(builtNADicts)
    print msg
    outputMessage.append(msg)

    count=1
    ncount=1
    for i in builtNADicts:
        if len(outfilenames)==1:
	    if len(builtNADicts)==1:
	        suffix=""
	    else:
	        suffix="_%s" % ncount
	    nameparts=outfilenames[0].split(".")    
	    newname=(".".join(nameparts[:-1]))+suffix+"."+nameparts[-1]
	else:
	    newname=outfilenames[count-1]
 
	msg="\nWriting output NASA Ames file: %s" % newname
	print msg
	outputMessage.append(msg)
	
	builtNADict=i[0]
	for key in naVars.keys():
	    if key in allowedOverwriteMetadata:
	    
	        if key in arrayArgs:
		    newItem=naVars[key].split()		   
		else:
	            newItem=naVars[key]
		    		    
		if newItem!=builtNADict[key]:
		    builtNADict[key]=newItem
		    msg="Metadata overwritten in output file: '%s' is now '%s'" % (key, builtNADict[key])
		    print msg
		    outputMessage.append(msg)
        
        fileList=[]
        # Cope with size limits if specified and FFI is 1001
        if sizeLimit and (builtNADict["FFI"]==1001 and len(builtNADict["V"][0])>sizeLimit):
            varList=builtNADict["V"]
            arrayLength=len(varList[0])
            nvolInfo=divmod(arrayLength, sizeLimit)
            nvol=nvolInfo[0]
            if nvolInfo[1]>0: nvol=nvol+1
            start=0
            letterCount=0
            ivol=0
            while start<arrayLength:
                ivol=ivol+1
                end=start+sizeLimit
                if end>arrayLength:
                    end=arrayLength
                currentBlock=[]
                # Write new V array
                for v in varList:
                    currentBlock.append(v[start:end])

                # Adjust X accordingly
                NADictCopy=modifyNADictCopy(builtNADict, currentBlock, start, end, ivol, nvol)
                
                # Write data to output file
                newnamePlusLetter="%s-%.3d.na" % (newname[:-3], ivol)
                fileList.append(newnamePlusLetter)
                general.openNAFile(newnamePlusLetter, 'w', NADictCopy, delimiter=delimiter, float_format=float_format)
                msg="\nOutput files split on size limit: %s\nFilename used: %s" % (sizeLimit, newnamePlusLetter)
                print msg
                outputMessage.append(msg)
                letterCount=letterCount+1
                start=end


        else:		
   	    general.openNAFile(newname, 'w', builtNADict, delimiter=delimiter, float_format=float_format)

	msg="\nWrote the following variables:"+"\n\t"+("\n\t".join(i[1][0]))
	print msg
	outputMessage.append(msg)
	
	if len(i[1][1])>0:
	    msg="\nWrote the following auxiliary variables:"
	    msg=msg+"\n\t"+("\n\t".join(i[1][1]))	
	    
	if len(i[1][2])>0:
	    msg="\nWrote the following Singleton variables:"
	    msg=msg+"\n\t"+("\n\t".join(i[1][2]))

        if len(fileList)>0:
            msg=msg+("\n\nNASA Ames files written successfully: \n%s" % "\n".join(fileList))
            count=count+len(fileList)
        else:
	    msg=msg+"\n\nNASA Ames file written successfully: %s" % newname
            count=count+1
        ncount=ncount+1

	print msg
	outputMessage.append(msg)
	    
    if (count-1)==1:
        plural=""
    else:
        plural="s"	      
    msg="\n%s file%s written." % ((count-1), plural)
    print msg
    outputMessage.append(msg)
    return outputMessage


class NAContentCollector(NACore):
    """
    Class to build a NASA Ames File object from a set of 
    CDMS variables and global attributes (optional).
    """
    
    def __init__(self, vars, global_attributes={}, cdmsfile=None, rule=None):
        """
        Sets up instance variables and calls appropriate methods to
        generate sections of NASA Ames file object.
        """
        self.rule=rule
        self.cdmsfile=cdmsfile
	self.outputMessage=[]
        self.na_dict={}
        self.vars=vars
        self.varIDs=None
        self.globals=global_attributes	
	self.rankZeroVars=[]
	self.rankZeroVarIDs=[]
        (self.orderedVars, auxVars)=self.analyseVariables()
	if self.orderedVars==None:
	    self.varBin=[]
	else:
	    #print "NAMELISTS:", [var.id for var in self.orderedVars],[var.id for var in auxVars]
	    self.varIDs=[[var.id for var in self.orderedVars],[var.id for var in auxVars], self.rankZeroVarIDs]
	
            self.na_dict["NLHEAD"]="-999"
	
	    #print [var.id for var in self.orderedVars]
	    #print [var.rank() for var in self.orderedVars]	
            self.defineNAVars(self.orderedVars)
            self.defineNAAuxVars(auxVars)
            self.defineNAGlobals()
            self.defineNAComments()
            self.defineGeneralHeader()
            # Quick fudge
            if self.na_dict["FFI"]==1001: self.na_dict["X"]=self.na_dict["X"][0]


    def analyseVariables(self):
        """
	Method to examine the content of CDMS variables to return
	a tuple of two lists containing variables and auxiliary variables
	for the NASA Ames file object.
	Variables not compatible with the first file are binned to be used next.
	"""
	# Need to group the variables together in bins
	self.varBin=[]
	# Get largest ranked variable as the one we use as standard
	highrank=-1
        bestVar=None
	count=0
	for var in self.vars:
	    msg="Analysing: %s" % var.id
	    print msg
	    self.outputMessage.append(msg)
	    count=count+1

	    # get rank
	    rank=var.rank()

            # Deal with specific datasets with special rules
            if self.rule!=None and self.rule[0]=="aircraft":
                var=self._useLocalRule(var, self.rule)
                if type(var)==type(None): 
                    continue 
                rank=1

            # Deal with singleton variables
	    if rank==0: 
	        self.rankZeroVars.append(var)
		self.rankZeroVarIDs.append(var.id)
		continue
	    
	    if rank>highrank:
                highrank=rank
		bestVar=var
		bestVarIndex=count
            elif rank==highrank:
	        if len(var.flat)>len(bestVar.flat):
		    bestVar=var
		    bestVarIndex=count
	
	if len(self.rankZeroVars)==len(self.vars):  return (None, None)
	if not bestVar:  
            print "No variables produced"
            return (None, None)

        vars4NA=[bestVar]
        auxVars4NA=[]
        shape=bestVar.shape
        ndims=len(shape)
        self.na_dict["NIV"]=ndims

        # Work out which File Format Index is appropriate 
        if ndims in (2,3,4):
            self.na_dict["FFI"]=10+(ndims*1000)
        elif ndims>4:
            raise "Cannot write variables defined against greater than 4 axes in NASA Ames format."
        else:
            if len(auxVars4NA)>0 or (self.na_dict.has_key("NAUXV") and self.na_dict["NAUXV"]>0):
                self.na_dict["FFI"]=1010
            else:
                self.na_dict["FFI"]=1001
        #print self.na_dict["FFI"]
        axes=bestVar.getAxisList()
        
        # Get other variable info
	#print [v.id for v in self.vars], bestVarIndex
	#print [v.id for v in self.vars[:bestVarIndex-1]+self.vars[bestVarIndex:]]
        for var in self.vars[:bestVarIndex-1]+self.vars[bestVarIndex:]:
            # Deal with specific datasets with special rules
            if self.rule!=None and self.rule[0]=="aircraft":
                if var.rank()==2:
                    var=self._useLocalRule(var, self.rule)
                    if type(var)==type(None): continue

	    #print self.rankZeroVars
	    #for rzv in self.rankZeroVars:  
	    #    if var.id==rzv.id and var[0]==rzv[0]: continue
	    #print [v.id for v in self.rankZeroVars]
	    if var.id in self.rankZeroVarIDs: continue
	    #print var.id, ndims, shape, len(var.shape), var.shape
            if len(var.shape)!=ndims or var.shape!=shape: 
                # Could it be an auxiliary variable 
                if len(var.shape)!=1: 
		    self.varBin.append(var)
		    continue
                caxis=var.getAxis(0)
                if compareAxes(axes[0], caxis)==0: 
		    self.varBin.append(var)
		    continue
                # I think it is an auxiliary variable
                auxVars4NA.append(var) 
		# Also put it in var bin because auxiliary vars might be useful
		self.varBin.append(var)
            else:
                caxes=var.getAxisList()
		#print var.id, "here"
                for i in range(ndims):            
                    if compareAxes(axes[i], caxes[i])==0:
		        self.varBin.append(var)
                        continue
                # OK, I think they are compatible
                vars4NA.append(var)
		
        # Re-order if they previously came from NASA Ames files (i.e. including 
	# the attribute 'nasa_ames_var_number')
	orderedVars=[None]*1000
	otherVars=[]
	for var in vars4NA:
	    if hasattr(var, "nasa_ames_var_number"):
	        orderedVars[var.nasa_ames_var_number[0]]=var
            else:
	        otherVars.append(var)
	# Remake vars4NA now in order
	vars4NA=[]
	for var in orderedVars:
	    if var!=None: vars4NA.append(var)
        vars4NA=vars4NA+otherVars

        # Now re-order the Auxiliary variables if they previously came from NASA 
	# Ames files (i.e. including the attribute 'nasa_ames_aux_var_number')

	orderedAuxVars=[None]*1000
	otherAuxVars=[]
	for var in auxVars4NA:
	    if hasattr(var, "nasa_ames_aux_var_number"):
	        orderedAuxVars[var.nasa_ames_aux_var_number[0]]=var
            else:
	        otherAuxVars.append(var)
	# Remake auxVars4NA now in order
	auxVars4NA=[]
	for var in orderedAuxVars:
	    if var!=None: auxVars4NA.append(var)
        auxVars4NA=auxVars4NA+otherAuxVars	
        return (vars4NA, auxVars4NA)


    def defineNAVars(self, vars):
        """
	Method to define NASA Ames file object variables and their
	associated metadata.
	"""
        self.na_dict["NV"]=len(vars)
        self.na_dict["VNAME"]=[]
        self.na_dict["VMISS"]=[]
        self.na_dict["VSCAL"]=[]
        self.na_dict["V"]=[]
        for var in vars:
            name=getBestName(var)
            self.na_dict["VNAME"].append(name)
            miss=getMissingValue(var)
            if type(miss) not in (float, int, long):  miss=miss[0]
            self.na_dict["VMISS"].append(miss)
            #print self.na_dict["VMISS"]
            self.na_dict["VSCAL"].append(1)
            # AND THE ARRAY
            # Populate the variable list  
            ######## NOTE - might not have to do this #####
            ######## It  might handle writing from a Numeric array ########
            self.na_dict["V"].append(var._data)
            #listOfListsCreator(inlist, var.shape)
            #arrayToList(var, inlist)

            if not self.na_dict.has_key("X"):
                self.na_dict["NXDEF"]=[]
                self.na_dict["NX"]=[]
                # Create independent variable information
		#print var.id, var.getAxis(0)
                self.ax0=var.getAxis(0)
                self.na_dict["X"]=[list(self.ax0._data_)]
                self.na_dict["XNAME"]=[getBestName(self.ax0)]
                if len(self.ax0)==1:
                    self.na_dict["DX"]=[0]
                else:
                    incr=self.ax0[1]-self.ax0[0]
		    # Set default increment as gap between first two
		    self.na_dict["DX"]=[incr]
		    # Now overwrite it as zero if non-uniform interval in axis
                    for i in range(1, len(self.ax0)):
                        if (self.ax0[i]-self.ax0[i-1])!=incr:
                            self.na_dict["DX"]=[0]
                            break

                # Now sort the rest of the axes
                for axis in var.getAxisList()[1:]:
                    self.getAxisDefinition(axis)


    def defineNAAuxVars(self, auxVars):
        """
	Method to define NASA Ames file object auxiliary variables and their
	associated metadata.
	"""
        self.na_dict["NAUXV"]=len(auxVars)
        self.na_dict["ANAME"]=[]
        self.na_dict["AMISS"]=[]
        self.na_dict["ASCAL"]=[]
        self.na_dict["A"]=[]
        for var in auxVars:
            name=getBestName(var)
            self.na_dict["ANAME"].append(name)
            miss=getMissingValue(var)
            if type(miss)!=float:  miss=miss[0]
            self.na_dict["AMISS"].append(miss)
            self.na_dict["ASCAL"].append(1)
            # AND THE ARRAY
            # Populate the variable list  
            ######## NOTE - might not have to do this #####
            ######## It  might handle writing from a Numeric array ########
            self.na_dict["A"].append(var._data)
            #listOfListsCreator(inlist, var.shape)
            #arrayToList(var, inlist)     


    def getAxisDefinition(self, axis):
        """
	Method to create the appropriate NASA Ames file object 
	items associated with an axis (independent variable in 
	NASA Ames).
	"""
	length=len(axis)
        self.na_dict["NX"].append(length)
        self.na_dict["XNAME"].append(getBestName(axis))
        # If only one item in axis values
        if length<2:
            self.na_dict["DX"].append(0)
            self.na_dict["NXDEF"].append(length)
            self.na_dict["X"].append(list(axis._data_))        
            return
   
        incr=axis[1]-axis[0]
        for i in range(1, length):
            if (axis[i]-axis[i-1])!=incr:
                self.na_dict["DX"].append(0)
                self.na_dict["NXDEF"].append(length)
                self.na_dict["X"].append(list(axis._data_))
                break
        else:
	    maxLength=length
	    if length>3: maxLength=3
            self.na_dict["DX"].append(incr)
            self.na_dict["NXDEF"].append(maxLength)
            self.na_dict["X"].append(axis[:maxLength])
        return


    def defineNAGlobals(self):
        """
	Maps CDMS (NetCDF) global attributes into NASA Ames Header fields.
	"""
	# Get the global mapping dictionary
        globalmap=cdmsMap.toNA
	# Check if we should add to it with locally set rules
	locGlobs=localRules.localGlobalAttributes
        for att in locGlobs.keys():
	    if not globalmap.has_key(att):
	        globalmap[key]=locGlobs[key]

        self.extra_comments=[[],[],[]]  # Normal comments, special comments, other comments
	conventionOrReferenceComments=[]
        for key in self.globals.keys():
            if key!="first_valid_date_of_data" and type(self.globals[key]) not in (str, float, int): continue
            if key in globalmap.keys():
                if key=="history":
                    timestring=time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
                    history="History:\t%s - Converted to NASA Ames format using nappy-%s.\n\t%s" % (timestring, version.version, self.globals[key])
                    history=history.split("\n") 
                    self.history=[]
                    for h in history:
                        if h[:8]!="History:" and h[:1]!="\t": h="\t"+h
                        self.history.append(h) 
                    
                elif key=="institution":
		    # If fields came from NA then extract appropriate fields.
		    match=re.match(r"(.*)\s+\(ONAME from NASA Ames file\);\s+(.*)\s+\(ORG from NASA Ames file\)\.", self.globals[key])
		    if match:
		        self.na_dict["ONAME"]=match.groups()[0]
			self.na_dict["ORG"]=match.groups()[1]
		    else:
                        self.na_dict["ONAME"]=self.globals[key]
                        self.na_dict["ORG"]=self.globals[key]		    
		    
		    # NOte: should probably do the following search and replace on all string lines
		    self.na_dict["ONAME"]=self.na_dict["ONAME"].replace("\n", "  ")
		    self.na_dict["ORG"]=self.na_dict["ORG"].replace("\n", "  ")
		    		    
                elif key=="comment":
		    # Need to work out if they are actually comments from NASA Ames in the first place
                    #self.ncom=[self.globals[key]]
		    comLines=self.globals[key].split("\n")
		    normComms=[]
		    normCommFlag=None
		    specComms=[]
		    specCommFlag=None
		    for line in comLines:
		        if line.find("###NASA Ames Special Comments follow###")>-1:
			    specCommFlag=1
			elif line.find("###NASA Ames Special Comments end###")>-1:
			    specCommFlag=None
		        elif line.find("###NASA Ames Normal Comments follow###")>-1:
			    normCommFlag=1
			elif line.find("###NASA Ames Normal Comments end###")>-1:
			    normCommFlag=None	
			elif specCommFlag==1:
			    specComms.append(line)
			elif normCommFlag==1:
			    normComms.append(line)
			elif line.find("###Data Section begins on the next line###")>-1:
			    pass
			else:
			    normComms.append(line)	    
		    
		    self.extra_comments=[specComms, normComms, []]		    
		   		    
		elif key=="first_valid_date_of_data":
		    self.na_dict["DATE"]=self.globals[key]
		
                elif key in ("Conventions", "references"):
                    #conventionOrReferenceComments.append("%s:   %s" % (key, self.globals[key]))
		    self.extra_comments[2].append("%s:   %s" % (key, self.globals[key]))
                else:
                    self.na_dict[globalmap[key]]=self.globals[key]
            else:
                self.extra_comments[2].append("%s:   %s" % (key, self.globals[key]))
	#self.extra_comments
        return


    def defineNAComments(self, normal_comments=[], special_comments=[]):
        """
	Defines the Special and Normal comments sections in the NASA Ames file 
	object - including information gathered from the defineNAGlobals method.
	"""
	
        if hasattr(self, "ncom"):  normal_comments=self.ncom+normal_comments
	NCOM=[]
        for ncom in normal_comments:
            NCOM.append(ncom)
        if len(NCOM)>0:   NCOM.append("")
	
	if len(self.extra_comments[2])>0:
	    for excom in self.extra_comments[2]:
	        NCOM.append(excom)
	
        if len(self.extra_comments[1])>0:  
	    NCOM.append("Additional Global Attributes defined in the source file and not translated elsewhere:")
            for excom in self.extra_comments[1]:
                NCOM.append(excom)

        if hasattr(self, "history"):
            for h in self.history:
                NCOM.append(h)
        
	if len(NCOM)>0:
	    NCOM.insert(0, "###NASA Ames Normal Comments follow###")
	    NCOM.append("")
	    NCOM.append("###NASA Ames Normal Comments end###")
        NCOM.append("###Data Section begins on the next line###")

        specCommentsFlag=None
	SCOM=[]
	special_comments=self.extra_comments[0]
	if len(special_comments)>0: 
	    SCOM=["###NASA Ames Special Comments follow###"]
	    specCommentsFlag=1
        for scom in special_comments:
            SCOM.append(scom)


        #used_var_atts=("name", "long_name", "standard_name", "id", 
	    #    "missing_value", "fill_value", "units", 
		#"nasa_ames_var_number", "nasa_ames_aux_var_number")
        used_var_atts=("id",  "missing_value", "fill_value", "units", 
                   "nasa_ames_var_number", "nasa_ames_aux_var_number")
        varCommentsFlag=None

        # Create a string for the Special comments to hold rank-zero vars
	rankZeroVarsString=[]
	for var in self.rankZeroVars:
	    rankZeroVarsString.append("\tVariable %s: %s" % (var.id, getBestName(var)))
	    for att in var.attributes.keys():
	        value=var.attributes[att]
		if type(value) in (str, float, int):
		    rankZeroVarsString.append("\t\t%s = %s" % (att, var.attributes[att]))
	    #print "VALUES", dir(var), var._data ; rankZeroVarsString.append("\t\tvalue = %s" % var._data)
	
	if len(rankZeroVarsString)>0:
	    rankZeroVarsString.insert(0, "###Singleton Variables defined in the source file follow###")
	    rankZeroVarsString.append("###Singleton Variables defined in the source file end###")

        for var in self.orderedVars:
            varflag="unused"
            name=getBestName(var)
            for scom,value in var.attributes.items():
                if type(value) in (type([]), type(Numeric.array([0]))) and len(value)==1:
                    value=value[0]
                if type(value) in (str, float, int) and scom not in used_var_atts:
                    if varflag=="unused":
                        if varCommentsFlag==None:
                            varCommentsFlag=1
			    if specCommentsFlag==None:
			        SCOM=["###NASA Ames Special Comments follow###"]+rankZeroVarsString
                            SCOM.append("Additional Variable Attributes defined in the source file and not translated elsewhere:")
                            SCOM.append("###Variable attributes from source (NetCDF) file follow###")
                        varflag="using" 
                        SCOM.append("\tVariable %s: %s" % (var.id, name))
                    SCOM.append("\t\t%s = %s" % (scom, value))

        if varCommentsFlag==1:  SCOM.append("###Variable attributes from source (NetCDF) file end###")
        if specCommentsFlag==1:
	    SCOM.append("###NASA Ames Special Comments end###")

        """used_var_atts=("name", "long_name", "standard_name", "id", "missing_value", "fill_value", "units")
        for var in self.vars:
            for scom,value in var.attributes.items():
                name=getBestName(var)
                if type(value) in (str, float, int) and scom not in used_var_atts:
                    SCOM.append("\t%s: %s - %s" % (name, scom, value))"""

        # Strip out empty lines (or returns)
	NCOM_cleaned=[]
	SCOM_cleaned=[]
	#hiddenNewLineCount1=0
	for c in NCOM:
	    if c.strip() not in ("", " ", "  "):
	        #hiddenNewLineCount1=hiddenNewLineCount1+c.count("\n")
		# Replace new lines within one attribute with a newline and tab so easier to read
		lines=c.split("\n")
		for line in lines:
		    if line!=lines[0]: line="\t"+line
		    NCOM_cleaned.append(line)
		
	#hiddenNewLineCount2=0	
	for c in SCOM:
	    if c.strip() not in ("", " ", "  "): 	        
	        #hiddenNewLineCount2=hiddenNewLineCount2+c.count("\n")
		# Replace new lines within one attribute with a newline and tab so easier to read
	        #c=c.replace("\n", "\n\t")
		#SCOM_cleaned.append(c)
		lines=c.split("\n")
		for line in lines:
		    if line!=lines[0]: line="\t"+line
		    SCOM_cleaned.append(line)
		    
        self.na_dict["NCOM"]=NCOM_cleaned
        self.na_dict["NNCOML"]=len(self.na_dict["NCOM"])#+hiddenNewLineCount1
        self.na_dict["SCOM"]=SCOM_cleaned
        self.na_dict["NSCOML"]=len(self.na_dict["SCOM"])#+hiddenNewLineCount2
        return


    def defineGeneralHeader(self, header_items={}):
        """
	Defines known header items and overwrites any with header_items 
	key/value pairs.
	"""
	# Check if DATE field previously known in NASA Ames file
	time_now=time.strftime("%Y %m %d", time.localtime(time.time())).split()
	if not self.na_dict.has_key("RDATE"):
	    self.na_dict["RDATE"]=time_now
	
        if self.ax0.isTime():
            # Get first date in list
	    try:
                (unit, start_date)=re.match("(\w+)\s+?since\s+?(\d+-\d+-\d+)", self.ax0.units).groups()            
                comptime=cdtime.s2c(start_date)
                first_day=comptime.add(self.na_dict["X"][0][0], getattr(cdtime, unit.capitalize()))
                self.na_dict["DATE"]=string.replace(str(first_day).split(" ")[0], "-", " ").split()
	    except:
	        msg="Nappy Warning: Could not get the first date in the file. You will need to manually edit the output file."
		print msg
		self.outputMessage.append(msg)
		self.na_dict["DATE"]=("DATE", "NOT", "KNOWN")
        else: 
            if not self.na_dict.has_key("DATE"):
	        msg="Nappy Warning: Could not get the first date in the file. You will need to manually edit the output file."
		print msg
		self.outputMessage.append(msg)
	        self.na_dict["DATE"]=("DATE", "NOT", "KNOWN")
        self.na_dict["IVOL"]=1
        self.na_dict["NVOL"]=1
        for key in header_items.keys():
             self.na_dict[key]=header_items[key]
        return


    def _useLocalRule(self, var, ruleArgs):
        """
        Applies some logic based on a local rule.
        """
        ruleName=ruleArgs[0]
        rule=ruleArgs
        if ruleName=="aircraft":
            # Fixes aircraft data 2D time axis, missing values and does sub-selection by default
            flagVar=None

            # return if variable is time
            if var.id=="Time":
                return None

            if len(rule)<2: rule.append("av")
            if len(rule)<3: rule.append("flag")
            if rule[2]=="flag":
                # Only use flag var for processing real variable
                if var.id.strip()[-4:]=="FLAG": 
                    print "Ignore flag: %s" % var.id
                    return None 

                flagID=var.id.strip()+"FLAG"
                try:
                    flagVar=self.cdmsfile(flagID)
                except:
                    raise "Cannot get flag variable for '%s'" % var.id
            elif rule[2]=="noflag":
                flagVar=None

            timeVar=self.cdmsfile("Time")
            import localRules.aircraftData
            ruleArgs=[rule[1]]+rule[3:]
            fixedVar=localRules.aircraftData.AircraftData(var, timeVar, flagVar, ruleArgs).subSampledVar
            return fixedVar
        else:
            raise "Rule '%s' not yet defined." % ruleName
                

usenc2nainstead="""if __name__=="__main__":

    args=sys.argv[1:]
    if len(args)<4:
        print helpMessage
        print "Incorrect number of arguments used."
        sys.exit()
	
    for arg in args:
        if arg=="-i":
	    infile=args[args.index(arg)+1]
	elif arg=="-o":
	    outfile=args[args.index(arg)+1]

    cdms2na(infile, outfile) """



"""
nappy_api.py
============

Top-level API module that allows user to access most of the useful stuff in
nappy. API examples:

 1. Working with NASA Ames file objects
 2. Converting between formats (NASA Ames, NetCDF and CSV)
 3. Comparing NASA Ames files (and/or CSV files)
 4. General NASA Ames utilities 

 1. Working with NASA Ames file objects

# Start python interactive shell
$ python

# Import the nappy package
import nappy

# Let's open a NASA Ames file and examine its contents
f = nappy.openNAFile("data_files/2010.na")

# Get number of header lines
n_lines = f.getNumHeaderLines()

# Get Organisation from header
org = f.getOrg()
# Get the Normal Comments (SCOM) lines.
norm_comms = f.getNormalComments()

# Get the Special Comments (SCOM) lines.
spec_comms = f.getSpecialComments()

# Get a list of metadata for all main (non-auxiliary or independent) variables
var_list = getVariables()

# Get Auxiliary variable metadata for auxiliary variable number 2
(variable, units, miss, scale) = f.getAuxVariable(2)

# Get scale factor for primary variable number 3
scale_factor = f.getScaleFactor(3)

# Get missing value for primary variable number 1
missing = f.getMissingValue(1)

# Let's get the contents dictionary of the whole file
na_dict = f.getNADict()

# Let's write the na_dict object to a new NASA Ames file
fout = openNAFile("test_outputs/mytest.na", mode="w", na_dict=na_dict)
fout.write()
fout.close()

 2. Converting between formats (NASA Ames, NetCDF and CSV)

# Let's convert a NASA Ames file into a NetCDF file, and add some of our own global attributes
glob_atts = [("Project", "Really important scientific project involving worms"),
             ("Errata": "I meant worm holes!")]
na_file = "data_files/1020.na"
nc_file = "test_outputs/try_1020.nc"
nappy.convertNAToNC(na_file, nc_file, global_attributes=glob_atts)

# Let's convert a NASA Ames file to a CSV and add an annotation column to explain the header
nappy.convertNAToCSV(na_file, annotation=True)

# Let's read a NetCDF and write one (or more) output NASA Ames files,
# but only including and variables "temp" and "ozone". Also let's write 
# the output using tabs as the delimiters and a float format of "%6.3f".
nappy.convertNCToNA("data_files/test1.nc", "test_outputs/test1nc.na", 
              var_ids=("temp", "ozone"), delimiter="\t", float_format="%6.3f")
  
# Let's convert a NetCDF file to one (or more) CSV files and don't write the header at all
nappy.convertNCToCSV("data_files/test1.nc", "test_outputs/test1nc_no_header.csv",
                     no_header=True)

# Let's take some in-memory CDMS objects and write them to one, or more, NASA Ames file(s).
# We need to give it a list of cdms variables and a global attributes list of tuples/lists.
# We also want to instruct nappy to overwrite the content of its 
# MNAME (Mission Name) header line with our specific mission name.
# Also, tell nappy to write the output to NASA Ames file format index (FFI) 2310
# because we know it is compatible.
nappy.convertCDMSObjectsToNA([cdms_var_1, cdms_var_2], [("Institute", "British Atmospheric Data Centre")], 
              na_file="test_outputs/cdms_to_na.na", 
              na_items_to_override={"MNAME": "Atlantic Divergence Mission 2009"}, 
              requested_ffi=2310)

# Let's take a list of cdms variables and a global attributes list and write
# them to a CSV file.
nappy.convertCDMSObjectsToCSV(cdms_vars, global_atttributes, csv_file)

# Let's take a NASA Ames dictionary object, and write it to a NetCDF file
nappy.writeNADictToNC(na_dict, nc_file, mode="w")

# Let's try and write a second na_dict object to the same NetCDF file using mode="a".
nappy.writeNADictToNC(na_dict_2, nc_file, mode="a")

# Now let's read in a NASA Ames file and convert the contents in-memory into
# CDMS objects so that we can manipulate them with NetCDF-compatible tools
(cdms_vars_primary, cdms_vars_aux, global_attributes) = nappy.readCDMSObjectsFromNA(na_file)

# Actually, I only want to get a single variable from that file, so I'll try
temp_var = getCDMSVariableFromNA(na_file, "temperature")

 3. Comparing NASA Ames files (and/or CSV files)

# I'd like to compare a NASA Ames and CSV file to check they are the same.
# It will allow for different formatting of numbers as long as the values 
# are the same. Compare both header and body by setting as True (default).
result = nappy.compareNA(na_file, csv_file, header=True, body=True, 
            number_clever=True, delimiter_1="    ", delimiter_2=",")

 4. General NASA Ames utilities 

# Get the FFI from a NASA Ames file
ffi = nappy.readFFI(na_file)

# Given a NASA Ames dictionary (na_dict) get an appropriate FFI.
ffi = nappy.chooseFFI(na_dict)


"""

# Import standard library modules
import logging
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

# Import third-party software
try:
    import cdms2 as cdms
except:
    try:
        import cdms
    except:
        log.warn("You cannot use nappy NetCDF conversion tools as your system does not have CDMS installed, or it is not in your sys.path.")
        cdms = False

# Import local modules
import nappy.utils.common_utils
import nappy.utils.compare_na
import nappy.utils.text_parser

# Bring some utils into the API
compareNA = nappy.utils.compare_na.compareNA
readFFI = nappy.utils.common_utils.readFFI
chooseFFI = nappy.utils.common_utils.chooseFFI
getNAFileClass = nappy.utils.common_utils.getNAFileClass
getFileNameWithNewExtension = nappy.utils.common_utils.getFileNameWithNewExtension

__version__ = nappy.utils.common_utils.getVersion()
default_delimiter = nappy.utils.common_utils.getDefault("default_delimiter")
default_float_format = nappy.utils.common_utils.getDefault("default_float_format")


def openNAFile(filename, mode="r", na_dict=None):
    """
    Function wrapper around the NASA Ames File classes. Any NASA Ames
    file can be opened through this function and the appropriate read or
    write NASA Ames File class instance is returned.
    """
    if mode == "r":
        ffi = readFFI(filename)
        return apply(getNAFileClass(ffi), (filename, mode))

    elif mode == "w":
        if na_dict.has_key('FFI') and type(na_dict['FFI']) == type(3):
            ffi = na_dict['FFI']
        else:
            ffi = chooseFFI(na_dict)
            na_dict['FFI'] = ffi
            log.info("\nFormat identified as: %s" % ffi)
        return apply(getNAFileClass(ffi), (filename,), {"mode":mode, "na_dict":na_dict})
    else:
        raise Exception("File mode not recognised '" + mode + "'.")


def convertNAToNC(na_file, nc_file=None, mode="w", variables=None, aux_variables=None,
                 global_attributes=[("Conventions", "CF-1.0")],
                 time_units=None, time_warning=True,
                 rename_variables={}):
    """
    Takes a NASA Ames file and converts to a NetCDF file. Options are:

    na_file - the input NASA Ames file.
    nc_file - name for the output NetCDF file (default is to replace ".na" from NASA Ames 
              file with ".nc").
    mode - is the file mode, either "w" for write or "a" for append
    variables - is a list of variable names that you wish to be converted. If not set then 
              nappy will attempt to convert all files.
    aux_var_list - is a list of auxiliary variables names that you wish to be converted. 
              If not set then nappy will use any compatible variables it finds as 
              auxiliary variables.
    global_attributes - is a list of global attributes to add to the output file.
    rename_variables - is a dictionary of {old_name: new_name} variable ID pairs that nappy 
              should use to rename variables before it writes them to file.  
    time_units - is a valid time units string such as "hours since 2003-04-30 10:00:00" to 
              use for time units if there is a valid time axis.
    time_warning - suppresses the time units warning for invalid time units if set to False.
    """
    arg_dict = vars()
    for arg_out in ("nc_file", "mode"):
        del arg_dict[arg_out]

    import nappy.nc_interface.na_to_nc
    convertor = apply(nappy.nc_interface.na_to_nc.NAToNC, [], arg_dict)
    convertor.convert()
    if nc_file == None:
        nc_file = getFileNameWithNewExtension(na_file, "nc")
    convertor.writeNCFile(nc_file, mode)
    return nc_file   
  

def convertNAToCSV(na_file, csv_file=None, annotation=False, no_header=False):
    """
    Reads in a NASA Ames file and writes it out a new CSV file which is identical to the
    input file except that commas are used as the delimiter. Arguments are:

    na_file - NASA Ames file path
    csv_file - CSV file path (default is to replace ".na" from NASA Ames file with ".csv").
    annotation - if set to True write the output file with an additional left-hand column 
                 describing the contents of each header line.
    no_header - if set to True then only the data blocks are written to file.
    """
    fin = openNAFile(na_file)
    fin.readData()
    na_dict = fin.getNADict()
    fin.close()

    if csv_file == None:
        csv_file = getFileNameWithNewExtension(na_file, "csv")

    fout = openNAFile(csv_file, "w", na_dict=na_dict)
    fout.write(delimiter=",", annotation=annotation)
    fout.close()
    return True


def convertNCToNA(nc_file, na_file=None, var_ids=None, na_items_to_override={},
            only_return_file_names=False, exclude_vars=[],
            requested_ffi=None, delimiter=default_delimiter, float_format=default_float_format, 
            size_limit=None, annotation=False, no_header=False,
            ):
    """
    Takes a NetCDF file and converts the contents to one or more NASA Ames files. 
    Arguments are:

    nc_file - is the name of input file (NetCDF).
    na_file - is the name of output file (default is to replace ".nc" from NASA Ames 
              file with ".na"). If multiple files produced then this name will be used 
              as the base name.
    var_ids - is a list of variables (as ids) to include in the output file(s).
    na_items_to_override - is a dictionary of {key: value} pairs to overwrite in output 
              files. Typically the keys are in:  
              ("DATE", "RDATE", "ANAME", "MNAME","ONAME", "ORG", "SNAME", "VNAME".)
    only_return_file_names - if set to True then only return a list of file names that 
              would be written (i.e. don't convert actual file).
    exclude_vars - is a list of variables (as ids) to exclude in the output file(s).
    requested_ffi - is the NASA Ames File Format Index (FFI) you wish to write to. Note 
              that there are only limited options available depending on the data 
              structures found.
    delimiter - the delimiter you wish to use between data items in the output file such 
              as "   ", "\t" or ",".
    float_format - a python formatting string such as "%s", "%g" or "%5.2f" used for 
              formatting floats when written to file.
    size_limit - if format FFI is 1001 then chop files up into size_limit rows of data.
    annotation - if set to True write the output file with an additional left-hand column 
              describing the contents of each header line.
    no_header - if set to True then only the data blocks are written to file.
    """
    arg_dict = vars()
    for arg_out in ("na_file", "only_return_file_names", "delimiter", "float_format", 
                    "size_limit", "annotation", "no_header"):
        del arg_dict[arg_out]

    if na_file == None:
        na_file =  getFileNameWithNewExtension(nc_file, "na")

    import nappy.nc_interface.nc_to_na
    convertor = apply(nappy.nc_interface.nc_to_na.NCToNA, [], arg_dict)
    convertor.convert()

    # If user only wants files then only give them that
    if only_return_file_names == True:
        return convertor.constructNAFileNames(na_file)
    else:
        convertor.writeNAFiles(na_file, delimiter=delimiter, float_format=float_format, 
                               size_limit=size_limit, annotation=annotation, no_header=no_header)
        log.info(convertor.output_message)
        output_files_written = convertor.output_files_written
        log.info(output_files_written)
        return output_files_written

    
def convertNCToCSV(nc_file, csv_file=None, **arg_dict):
    """
    Reads in a NetCDF file and writes the data out to a CSV file following the
    NASA Ames standard.
    """
    if csv_file == None:
        csv_file = getFileNameWithNewExtension(nc_file, "csv")
        arg_dict["na_file"] = csv_file
        arg_dict["delimiter"] = ","
  
    return apply(convertNCToNA, [nc_file], arg_dict)
    

def convertCDMSObjectsToNA(cdms_vars, global_attributes, na_file, 
              na_items_to_override={}, requested_ffi=None, delimiter=default_delimiter, 
              float_format=default_float_format, size_limit=None, annotation=False, no_header=False,
              ):
    """
    Takes a list of cdms variables and a list of global attributes and
    writes them to one or more NASA Ames files. Arguments are:
 
    cdms_vars - is a list of CDMS variables
    global_attributes - is a list of (key, value) pairs for header
    na_file - is the name of output file. If multiple files produced then this name will be used
              as the base name.
    na_items_to_override - is a dictionary of {key: value} pairs to overwrite in 
                output files. Typically the keys are in:  
                ("DATE", "RDATE", "ANAME", "MNAME","ONAME", "ORG", "SNAME", "VNAME", "SCOM", "NCOM".)
    requested_ffi - is the NASA Ames File Format Index (FFI) you wish to write to. 
                Note that there are only limited options available depending on the data 
                structures found.
    delimiter - the delimiter you wish to use between data items in the output file 
                such as "   ", "\t" or ",".
    float_format - a python formatting string such as "%s", "%g" or "%5.2f" used for 
                formatting floats when written to file.
    size_limit - if format FFI is 1001 then chop files up into size_limit rows of data.
    annotation - if set to True write the output file with an additional left-hand 
                column describing the contents of each header line.
    no_header - if set to True then only the data blocks are written to file.
    """
    import nappy.nc_interface.cdms_objs_to_na_file
    convertor = nappy.nc_interface.cdms_objs_to_na_file.CDMSObjectsToNAFile(cdms_vars, global_attributes=global_attributes, 
                        na_items_to_override=na_items_to_override, requested_ffi=requested_ffi,
                        )
    convertor.convert()
    na_files = convertor.writeNAFiles(na_file, delimiter=delimiter, float_format=float_format, 
                                      annotation=annotation, no_header=no_header) 
    return convertor.output_files_written 


def convertCDMSObjectsToCSV(cdms_vars, global_attributes, csv_file, **arg_dict):
    """
    Takes a list of cdms variables and a global attributes list and
    writes them to one or more CSV files.
    """
    arg_dict["delimiter"] = ","
    return apply(convertCDMSObjectsToNA, [cdms_vars, global_attributes, csv_file], arg_dict)


def writeNADictToNC(na_dict, nc_file, mode="w"):
    """
    Writes an NASA Ames dictionary object called na_dict to a NetCDF file called nc_file.
    Can set mode="a" or mode="w" to either append to existing nc_file or write new one.
    Note that mode="a" might not always work.
    """
    # Note, this needs to pretend that the na_dict exists, do this by instantiating NACore and cheating...
    import nappy.na_file.na_core
    na_file_obj = nappy.na_file.na_core.NACore()
    na_file_obj.setNADict(na_dict)
    # Fake up some required methods
    def fakeCaller():pass
    na_file_obj.readData = fakeCaller
    import nappy.nc_interface.na_to_cdms
    convertor = nappy.nc_interface.na_to_cdms.NAToCDMS(na_file_obj)
    (cdms_primary_vars, cdms_aux_vars, global_attributes) = convertor.convert()

    # Now write them out
    fout = cdms.open(nc_file, mode=mode)
    for var in (cdms_primary_vars + cdms_aux_vars):
        fout.write(var)

    # Write global attributes
    for (att, value) in global_attributes:
        setattr(fout, att, value)
        
    fout.close()
    log.info("NetCDF file '%s' written successfully." % file_name)
    return True


def readCDMSObjectsFromNA(na_file):
    """
    Reads the NASA Ames file and converts to CDMS objects.
    Returns a tuple containing:
      * a list of primary NASA Ames variables as CDMS variables
      * a list of auxiliary NASA Ames variables as CDMS variables,
      * a list of global attributes
    """
    cdms_var_list = []
    global_attributes = {}

    # Open the NA file
    na_file_obj = openNAFile(na_file)
    import nappy.nc_interface.na_to_cdms
    convertor = nappy.nc_interface.na_to_cdms.NADictToCdmsObjects(na_file_obj)
    (cdms_vars_primary, cdms_vars_aux, global_attributes) = convertor.convert()
    return (cdms_vars_primary, cdms_vars_aux, global_attributes)


def getCDMSVariableFromNA(na_file, var):
    """
    Returns a CDMS variable object (TransientVariable) identified by the var argument which
    can either be an integer index in the list of variables or the name of the variable.
    The variable is created from the variables found in the NASA Ames file na_file.
    """
    na_file_obj = openNAFile(na_file)
    import nappy.nc_interface.na_to_cdms
    convertor = nappy.nc_interface.na_to_cdms.NADictToCdmsObjects(na_file_obj, variables=[var])
    (cdms_primary_vars, cdms_aux_vars, global_attributes) = convertor.convert()
    # Must now be a primary var
    return cdms_primary_vars[0]





#!/usr/bin/env python

"""
compare_na.py
=============

Tool to compare contents of NASA Ames files or directories full of files.
Allows you to compare headers and data blocks in NASA Ames.

Usage:
======

    compare_na.py [-h | --header-only]  [-b | --body-only]
                  [-n | --number-strict] [-a | --approx-equal] 
                  [-1 <delimiter_1> | --delimiter-1=<delimiter_1>]
                  [-2 <delimiter_2> | --delimiter-2=<delimiter_2>]
                  <item1> <item2>
                                                                   

Where:
======

    <item1> and <item2> 	can either be a text file or directory.
    -h | --header-only    	selects compare only header(s)
    -b | --body-only		selects compare only body(s)
    -n | --number-strict	compares exact formatting of numbers in data block
                                (default is to compare them by value).
    -a | --approx-equal         considers any two numbers being compared the same as long
                                as the difference between them is less than 1%.
    <delimiter_1>		delimiter to use for file 1.
    <delimiter_2>		delimiter to use for file 2. 

"""

# Import standard library modules
import os
import sys
import re
import getopt

# Import local modules
from compare import *

equality_threshold = 0.01 # i.e. within 1% of each other
file_exclusion_patterns = (".*CSV.*", ".*svn.*", "\..*", ".*\.pyc$", ".*~$") 
file_exclusions = [re.compile(pattn) for pattn in file_exclusion_patterns]
letter_match = re.compile("[a-zA-Z]")


def exitNicely(msg):
    "Tidy exit."
    print __doc__
    print msg
    sys.exit()


def compareNA(i1, i2, **kwargs):
    """
    Compares items whether files or directories.
    Reports any differences at the command line but
    also returns them in a dictionary as:
    ???
    **kwargs are forwarded as dictionary to compNAFiles().
    """
    if os.path.isfile(i1):
        apply(compNAFiles, (i1, i2), kwargs)
    elif os.path.isdir(i1):
        compDirs(i1, i2)
    else:
        exitNicely("Cannot recognise/find item '" + i1 + "'.")


def compareSections(l1, l2, number_clever=True, approx_equal=False, 
                   delimiter_1=None, delimiter_2=None):
    """
    Compares sections of NASA Ames files (i.e. headers and bodies).
    """ 
    leng = len(l1)
    if len(l2) < leng:
        leng = len(l2)

    all_same = True

    for i in range(leng):

        # Start by setting same equal to True and then try and disprove this
        same = True

        l1[i] = l1[i].strip()
        l2[i] = l2[i].strip()

        # If letters found in line then not going to be numeric 
        # Hence we can just test if lines are identical
        if letter_match.search(l1[i]):
            if l1[i] == l2[i]:
                continue

        items1 = l1[i].split(delimiter_1)
        items2 = l2[i].split(delimiter_2)
        
        if len(items1) != len(items2):
            # Check that space delimiter hasn't just split identical lines to different lengths
            if len(items1) == 1 and items1[0].split() == items2:
                continue
            elif len(items2) == 1 and items2[0].split() == items1:
                continue
            else:
                same = False
        else: 
            if number_clever == False:
                if items1 != items2:
                    same = False
            else:
                for count in range(len(items1)):
                    try:
                        a = float(items1[count])
                        b = float(items2[count])
                    except:
                        a = items1[count]
                        b = items2[count]

                    if a != b:
                        
                        if approx_equal:
                        # Check to see if testing for approximate equality
                            if a == 0:  a = 0.000000001
                            if b == 0:  b = 0.000000001
                            divided = a/b
                            if divided < 1:
                                divided = b/a
                            if (1 - divided) > equality_threshold:
                                same = False
                                break  
                        else: 
                            same = False
                            break 
              
        if same == False:
            all_same = False
            print "Line %s:" % (i+1)
            print ">>>", l1[i]
            print "<<<", l2[i]

    return all_same


def compNAFiles(f1, f2, header=True, body=True, number_clever=True, approx_equal=False,
                delimiter_1=None, delimiter_2=None):
    """
    Compares contents of two NASA Ames files f1 and f2.
    header=False or body=False will not compare these sections of the files.
    number_clever=True will compare 5.00000 and 5 making them equal in the body.
    If approx_equal is True then approximate equality is good enough to return two
    numbers as being equal (within equality_threshold set at top of this module).
    If f1_delimiter and f2_delimiter are provided then the comparer will consider
    two lines identical if they have the delimiters sent in as arguments.
    """	
    name = os.path.split(f1)[-1]
    # Ignore anything that is in exclusion list
    for excl in file_exclusions:
        if excl.match(name):
            print "IGNORING EXCLUDED file:", f1
            return

    # Check they exist
    for f in (f1, f2):
        if not os.path.isfile(f):
            exitNicely("CANNOT compare files as item does not exist:" + f)
    
    # Note delimiter set as None will do split on white-space (which we want!)

    l1 = open(f1).readlines()
    l2 = open(f2).readlines()
    head_len1 = int(l1[0].split(delimiter_1)[0])
    head_len2 = int(l2[0].split(delimiter_2)[0])

    header1 = l1[:head_len1]
    header2 = l2[:head_len2]
    body1 = l1[head_len1:]
    body2 = l2[head_len2:]

    same = True
    if header == True:
        print "Comparing headers:"
        print ">>> %s header:" % f1
        print "<<< %s header:" % f2
        same = compareSections(header1, header2, number_clever, approx_equal, delimiter_1, delimiter_2) 
        if same == True:
            print "HEADERS ARE IDENTICAL."
        if len(header1) != len(header2):
            print "Header lengths differ:\n>>> %s: %s\n<<< %s: %s" % (f1, len(header1), f2, len(header2))

    if body == True:
        print "Comparing bodies:"
        print ">>> %s body:" % f1
        print "<<< %s body:" % f2
        same = compareSections(body1, body2, number_clever, approx_equal, delimiter_1, delimiter_2)
        if same == True:
            print "BODIES ARE IDENTICAL."
        if len(body1) != len(body2):
            print "Body lengths differ:\n>>> %s: %s\n<<< %s: %s" % (f1, len(body1), f2, len(body2))
       
    return same


def parseArgs(args):
    """
    Parses arguments returning a dictionary.
    """
    arg_dict = {}
    a = arg_dict
    a["header"] = True
    a["body"] = True
    a["number_clever"] = True
    a["approx_equal"] = False
    a["delimiter_1"] = None
    a["delimiter_2"] = None

    (arg_list, files) = getopt.getopt(args, "hbna1:2:", ["header-only", "body-only",
                    "number-strict", "approx-equal", "delimiter-1=", "delimiter-2="])

    for arg, value in arg_list:
        if arg in ("--header-only", "-h"):
            a["body"] = False
        elif arg in ("--body-only", "-b"):
            a["header"] = False 
        elif arg in ("--number-strict", "-n"):
            a["number_clever"] = False
        elif arg in ("--approx-equal", "-a"):
            a["approx_equal"] = True
        elif arg in ("--delimiter-1", "-1"):
            a["delimiter_1"] = value
        elif arg in ("--delimiter-2", "-2"):
            a["delimiter_2"] = value
        else:
            exitNicely("Unrecognised argument provided: " + arg)

    if len(files) != 2:
        exitNicely("Must provide a minimum of two file names as command line arguments.")

    if a["header"] == False and a["body"] == False:
        exitNicely("Invalid selection: header-only and body-only cannot be selected together.")

    return (files, a)


def main(args):
    "Main controller."
    files, arg_dict = parseArgs(args)
    apply(compareNA, files, arg_dict) 
   
 
if __name__=="__main__":

    args = sys.argv[1:]
    main(args)



#   Copyright (C) 2004 CCLRC & NERC( Natural Environment Research Council ).
#   This software may be distributed under the terms of the
#   Q Public License, version 1.0 or later. http://ndg.nerc.ac.uk/public_docs/QPublic_license.txt

"""
naFile2010.py
=============

Container module for NAFile2010 class.

"""

# Imports from python standard library

# Imports from local package
import nappy.utils.text_parser
import nappy.utils.list_manipulator
import nappy.na_file.na_file
import nappy.utils.common_utils
wrapLine = nappy.utils.common_utils.annotateLine
wrapLines = nappy.utils.common_utils.annotateLines


class NAFile2010(nappy.na_file.na_file.NAFile):
    """
    Class to read, write and interact with NASA Ames files conforming to the
    File Format Index (FFI) 2010.
    """

    def readHeader(self):
        """
        Reads FFI-specifc header section.
        """        
        self._normalized_X = False
        self._readCommonHeader()
        self.DX = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), self.NIV, float)
        self.DX.reverse()  # Reverse because C-type array is least-changing first
        self.NX = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), self.NIV - 1, int)
        self.NX.reverse()  # Reverse because C-type array is least - changing first
        self.NXDEF = nappy.utils.text_parser.readItemsFromLine(self.file.readline(), self.NIV - 1, int)
        self.NXDEF.reverse()  # Reverse because C-type array is least-changing first
        self.X = []
        for i in range(self.NIV - 1):
            self.X.append(nappy.utils.text_parser.readItemsFromUnknownLines(self.file, self.NXDEF[i], float))
        # Unbounded Independent variable should be first so insert empty list at start
        self.X.reverse()                 
        self.X.insert(0, [])
        self.XNAME = nappy.utils.text_parser.readItemsFromLines(self._readLines(self.NIV), self.NIV, str)
        self.XNAME.reverse()  # Reverse because C-type array is least-changing first
        self._readVariablesHeaderSection()
        self._readAuxVariablesHeaderSection()
        self._readComments()

    def writeHeader(self):
        """
        Writes FFI-specific header section.
        """        
        self._writeCommonHeader()
        DX = self.DX
        DX.reverse()
        self.header.write(wrapLine("DX", self.annotation, self.delimiter, (("%s" + self.delimiter) * (self.NIV - 1) + "%s\n") % tuple(DX)))
        NX = self.NX
        NX.reverse()
        self.header.write(wrapLine("NX", self.annotation, self.delimiter, (("%s" + self.delimiter) * (self.NIV - 2) + "%s\n") % tuple(NX)))
        NXDEF = self.NXDEF[:]
        NXDEF.reverse()
        self.header.write(wrapLine("NXDEF", self.annotation, self.delimiter, (("%s" + self.delimiter) * (self.NIV - 2) + "%s\n") % tuple(NXDEF)))

        X_lines = []

        for i in range(self.NIV - 1):
            X_lines.append((self.format * self.NXDEF[i] % tuple(self.X[i + 1][0:self.NXDEF[i]])).rstrip(" ,") + "\n")

        X_lines.reverse()
        for line in X_lines:
            self.header.write(wrapLine("X", self.annotation, self.delimiter, line.lstrip()))

        XNAME = self.XNAME[:]
        XNAME.reverse()
        self.header.write(wrapLines("XNAME", self.annotation, self.delimiter, "%s\n" * self.NIV % tuple(XNAME)))
        self._writeVariablesHeaderSection()
        self._writeAuxVariablesHeaderSection()
        self._writeComments()         
        self._fixHeaderLength()
        self.file.write(self.header.read())


    def _setupArrays(self):
        """
        Sets up FFI-specific arrays to fill with data (lists of lists).
        """
        self.V = []
        self.A = []

        # Create an array size to request using read routines
        self.arraySize = 1
        for i in self.NX:
        	self.arraySize = self.arraySize * i
        for n in range(self.NV):
            self.V.append([])
        for a in range(self.NAUXV):
            self.A.append([])
            
    def _readData1(self, datalines, ivar_count):
        """
        Reads first line/section of current block of data.
        """        
        # Start with independent and Auxilliary vars
        (x2_and_a, rtlines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, 1 + self.NAUXV, float)
        (x, aux) = (x2_and_a[0], x2_and_a[1:])
        self.X[0].append(x)
        count = 0
        for a in range(self.NAUXV):
            self.A[a].append(aux[count])
            count = count + 1
        return rtlines

    def _readData2(self, datalines, ivar_count):
        """
        Reads second line/section (if used) of current block of data.
        """
        # Now get the dependent variables
        for n in range(self.NV):
            (v, rtlines) = nappy.utils.text_parser.readItemsFromUnknownLines(datalines, self.arraySize, float)
            self.V[n].append([])
            nappy.utils.list_manipulator.recursiveListPopulator(self.V[n][ivar_count], v, self.NX)
            datalines = rtlines
        return rtlines

    def writeData(self):
        """
        Writes the data section of the file.
        This method can be called directly by the user.
        """        
        # Set up unbounded IV loop
        self.NX.reverse()
        for m in range(len(self.X[0])):

            # Write unbounded independent variable mark and auxiliary variables
            var_string = self.format % self.X[0][m]

            for a in range(self.NAUXV):
                var_string = var_string + (self.format % self.A[a][m])
            self.file.write(wrapLine("Data", self.annotation, self.delimiter, "%s\n" % var_string.rstrip(" ,")))
            
            # Write Variables
            for n in range(self.NV):
                outlines = nappy.utils.list_manipulator.recursiveListWriter(self.V[n][m], self.NX, delimiter = self.delimiter, float_format = self.float_format)
                for line in outlines:
                    self.file.write(wrapLine("Data", self.annotation, self.delimiter, line))

    def _normalizeIndVars(self):
        """
        Normalizes the values in the unbounded independent variable for FFIs
        that store an abbreviated version of this axis.
        """
        for i in range(self.NIV - 1):
            if self.NXDEF[i] == self.NX[i]:
                pass
            else:
                del self.X[i + 1][1:]
                count = 0
                while len(self.X[i + 1])<self.NX[i]:
                    nextx = self.X[i + 1][count] + self.DX[i + 1]
                    self.X[i + 1].append(nextx)
                    count = count + 1
        self._normalized_X = True
